// Copyright (C) 2020-2023 David Sugar, Tycho Softworks
// This code is licensed under MIT license

#ifndef FILESYSTEM_HPP_
#define FILESYSTEM_HPP_

#if __has_include(<filesystem>)
#include <filesystem>
namespace fsys = std::filesystem; // NOLINT
#else
#include <experimental/filesystem>
namespace fsys = std::experimental::filesystem; // NOLINT
#endif

#include <functional>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#if defined(__MINGW32__) || defined(__MINGW64__) || defined(WIN32)
extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
}
#endif

namespace tycho {
inline auto scan_stream(std::istream& input, std::function<bool(const std::string&)> proc) {
    std::string line;
    size_t count{0};
    while(std::getline(input, line)) {
        if (!proc(line))
            break;
        ++count;
    }
    return count;
}

inline auto scan_file(const fsys::path& path, std::function<bool(const std::string&)> proc) {
    std::fstream input(path);
    std::string line;
    size_t count{0};
    while(std::getline(input, line)) {
        if (!proc(line))
            break;
        ++count;
    }
    return count;
}

inline auto scan_directory(const fsys::path& path, std::function<bool(const fsys::directory_entry&)> proc) {
    auto dir = fsys::directory_iterator(path);
    return std::count_if(begin(dir), end(dir), proc);
}

inline auto scan_recursive(const fsys::path& path, std::function<bool(const fsys::directory_entry&)> proc) {
    auto dir = fsys::recursive_directory_iterator(path);
    return std::count_if(begin(dir), end(dir), proc);
}
} // end namespace

/*!
 * Filesystem support.
 * \file filesystem.hpp
 */
#endif
